﻿using System.Collections.Generic;
using System.Linq;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;

namespace GridControlApp
{
    public class GridControl : ItemsControl
    {
        public GridControl()
        {
            this.DefaultStyleKey = typeof(GridControl);
        }

        protected override Size MeasureOverride(Size availableSize)
        {
            base.MeasureOverride(availableSize);
            var all = SelfAndDescendants(this);
            foreach (DependencyObject d in all)
            {
                ((UIElement)d).InvalidateMeasure();
            }
            Grid itemsPanel = all
                .Where(x => x is Grid)
                .FirstOrDefault() as Grid;
            int rowsCount = itemsPanel.Children.Count;
            List<Panel> contentPanels = itemsPanel.Children
                .Select(presenter => VisualTreeHelper.GetChild(presenter, 0))
                .Select(panel => (Panel)panel)
                .ToList();
            int columnsCount = 0;
            foreach(Panel p in contentPanels)
            {
                if (columnsCount < p.Children.Count)
                {
                    columnsCount = p.Children.Count;
                }
            }
            itemsPanel.Children.Clear();
            for (int i = 0; i < rowsCount; i++)
            {
                itemsPanel.RowDefinitions.Add(new RowDefinition { Height = GridLength.Auto });
            }
            for (int i = 0; i < columnsCount; i++)
            {
                itemsPanel.ColumnDefinitions.Add(new ColumnDefinition { Width = GridLength.Auto });
            }
            for (int y = 0; y < rowsCount; y++)
            {
                for (int x = 0; contentPanels[y].Children.Count > 0; x++)
                {
                    FrameworkElement element = contentPanels[y].Children[0] as FrameworkElement;
                    contentPanels[y].Children.RemoveAt(0);
                    element.SetValue(Grid.ColumnProperty, x);
                    element.SetValue(Grid.RowProperty, y);
                    element.DataContext = contentPanels[y].DataContext;
                    itemsPanel.Children.Add(element);
                }
            }
            return base.MeasureOverride(availableSize);
        }

        private static IEnumerable<DependencyObject> SelfAndDescendants(DependencyObject d)
        {
            if (d == null) yield break;
            yield return d;
            int count = VisualTreeHelper.GetChildrenCount(d);
            for (int i = 0; i < count; i++)
            {
                DependencyObject child = VisualTreeHelper.GetChild(d, i);
                foreach (DependencyObject subitem in SelfAndDescendants(child))
                {
                    yield return subitem;
                }
            }
        }
    }
}
